/* Main routine is at the bottom of the file */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <signal.h>

#include "t_struct.h"
#include "t_format.h"
#include "t_parser.h"
#include "tsipincl.h"
#include "tsip_sio.h"
static int
	repeat_report_code,  /* report to be repeat-requested */
	new_repeat_report,
	nprintchar,				/* character count of ASCII line print */
	reply [0x200],			/* indexed by rpt.code; >0 if requested by cmd */
	serial_port_hold;

/* file for recording report data */
static FILE
	*tsipstream = NULL;

/* file name for same */
#define OUTBUFSIZE (25*80)
static char
	tsipfilename[80],
   outbuf[OUTBUFSIZE];
PORT_DEF
	pdef;

short
	tsipchat;	/* used in T_COMMND.C */

void SendTextToOutputDevice (char *obuf)
{
	char
   	*obufptr,
      *obufstart;
   obufptr=obuf;
   /* strip leading \n */
   if (*obuf == '\n') obufptr++;
   while (obufptr && *obufptr)
   {
      /* line feed, clear line, & print */
	   cputs ("\r\n");
		clreol();
   	obufstart = obufptr;
      /* look for next \n */
		obufptr = strchr (obufstart, '\n');
      /* if \n exists, create substring */
      if (obufptr)
      {
      	*obufptr = '\0';
      	obufptr++;
      }
		cputs (obufstart);
   }
}

/* serial port interface 'primitives' required by TSIP_IFC.C */
int getb (void)
{
	int this_byte;

	this_byte = get_char (pdef.port);
	if (this_byte != -1)
   {
		if (tsipstream != NULL) fputc (this_byte, tsipstream);
   }
	return this_byte;
}

int sendb (unsigned char outbyte)
{
	return put_char(pdef.port, outbyte);
}

/*
 * three special functions of TSIPCHAT:
 * repeat reports;
 * print non-TSIP characters;
 * hourly file opener.
*/

static void rawprint (int newbyte)
{
	int
		curr_window,
		newlinechar,
		newchar;

	switch (newbyte)
	{
	case '\r': case '\n':
		newlinechar = TRUE;
		newchar = -1;
		break;
	case '$': case '>':
	/* NMEA, TAIP starting characters */
		newlinechar = TRUE;
		newchar = newbyte;
		nprintchar++;
		break;
	default:
		newlinechar = FALSE;
		if (newbyte < 0x20 || newbyte > 0xFF) newchar = ((newchar&0x3F)|0x40);
		else newchar = newbyte;
		nprintchar++;
		break;
	}

	/* Save to reset screen at exit. */
	curr_window = which_window();
	switch_window (AUTOWIN);
	if (newlinechar || nprintchar > 75)
	{
		if (nprintchar) show_crlf();
		nprintchar = 0;
	}
	if (newchar > 0)
   {
		cprintf ("%c", (char)newchar);
	}
	switch_window (curr_window);
}

static void DisplayTSIPReport (TSIPPKT *rpt)
{
	int
		ireply;
   char
   	*obufptr;

   memset (outbuf, '\0', sizeof (outbuf));
   obufptr = outbuf;

	/* end of packet received, prepare for display */
	ireply = reply_code (rpt);
	if (reply [ireply])
	{
		/*
		** if this report has been requested, print in command window,
		** and decrement the reply request.  There may be 32 requests,
		** one for each SV: e.g., cmd_0x3B.
		*/
      if (repeat_report_code == 0x120) new_repeat_report = TRUE;
		if (ireply==repeat_report_code && new_repeat_report)
      {
			new_cmd_window ();
			/* the following will re-print the header if necessary */
			obufptr += print_msg_table_header (repeat_report_code, obufptr, TRUE);
			new_repeat_report = FALSE;
      }
		else
		{
			switch_window (CMDWIN);
      }
      reply [ireply]--;
   }
	else
	{
		switch_window (AUTOWIN);
   }

   /* Call T_REPORT.C routines and create a text report of the TSIP packet. */
	TranslateTSIPReportToText (rpt, obufptr);

   /* Display text report of the TSIP packet to the appropriate place(s). */
   SendTextToOutputDevice (outbuf);

	/* refresh repeat code for special display */
   reply [repeat_report_code] = 1;

   /* reset non-TSIP character printer */
   nprintchar = 0;

   /* prevent printing this packet again */
   rpt->status = TSIP_PARSED_EMPTY;

}

static void timekeeper (void)
{
	clock_t
		now_clock;
	static clock_t
		last_repeat_clock;
   static TSIPPKT
   	cmd;

 	now_clock = clock();
 	if (now_clock < last_repeat_clock + 1*CLK_TCK) return;

	if (!repeat_report_code) return;

	reply [repeat_report_code] = 1;
	switch (repeat_report_code)
	{
	case 0x41:
		cmd_0x21(&cmd);
		break;
	case 0x47:
		cmd_0x27(&cmd);
		break;
	case 0x4D:
		cmd_0x2D(&cmd);
		break;
	case 0x5A:
		cmd_0x3A(&cmd, 0);
		break;
 	case 0x5C:
		cmd_0x3C(&cmd, 0);
		break;
   case 0x120:
 		last_repeat_clock = now_clock;
   	return;
	default:
		return;
	}
	send_cmd(&cmd, sendb);
   new_repeat_report = TRUE;
 	last_repeat_clock = now_clock;
}

/*
 * The following routine is used in main and any routines requiring
 * user keyboard input.
 * By using this routine, the incoming serial stream is serviced and output on
 * the screen but can be pre-empted by processing of user keystrokes
 * while inputting data.
 */
int ProcessInputBytesWhileWaitingForKBHit (void)
{
	static TSIPPKT
		rpt;		      /* structure for current incoming TSIP report  */
	int
   	newbyte;
	int
		curr_window;   /* if in dual window configuration, current window */
   int
   	newkbch;

	/* Current screen state so we are able to reset screen at exit. */
	curr_window = which_window();

	/* do parsing until a kbhit is detected */
	while (!kbhit ()) {
		if (serial_port_hold)
		{
			/* suppress output if the serial port parmeters are being changed */
			continue;
		}

		newbyte = getb();
      if (newbyte >= 0)
      {
			tsip_input_proc (&rpt, (unsigned char)newbyte);
			if (rpt.status == TSIP_PARSED_FULL)
			{
      		DisplayTSIPReport (&rpt);
				/* return to active window if necessary */
				switch_window (curr_window);
			}
			else if (rpt.status == TSIP_PARSED_EMPTY && newbyte>=0)
			{
				/* non-TSIP character printer */
				rawprint (newbyte);
			}
      }

		/* repeat command alarm clock control every two seconds */
      timekeeper();
	}
	repeat_report_code = 0;
   newkbch = getch();
   if (!newkbch)
   {
   	newkbch = (char)(getch())*256;	/* extended char */
   }
	return newkbch;
}

static void ctrlchandler (short sig)
{
	exit (sig);
}

static void ProcessChatCommand (int kbch)
{
	short
		irpt;
	KB_CMD
		kb_cmd;
   TSIPPKT
   	cmd;

	repeat_report_code = 0;

	/* zero out all previous report requests */
	for (irpt=0; irpt < 0x200; irpt++) reply [irpt] = 0;

	/* open up command window */
	new_cmd_window ();

	/* 0x06, 0x07, 0x0C, 0x0D do not generate TSIP commands; */
	/* service them separately. */
	switch (kbch) {
	case 0x06:
		/* '^F' - continuous binary file storage control */
		if (tsipstream == NULL)
		{
			open_file_storage (&tsipstream, tsipfilename);
		}
		else
		{
			close_file_storage (&tsipstream, tsipfilename);
		}
		break;

	case 0x09:
		/* '^I' - change serial port parameters.
		/* When changing serial port settings, suppress TSIP parsing. */
		serial_port_hold = TRUE;
		if (!set_buf_serial_port (&pdef))
      {
/*      	setport (&pdef);*/
			if (ChangeSIOParms (&pdef))
         {
         	printf ("\nSubsequent set port failed");
            exit(0);
         }
      }
		serial_port_hold = FALSE;
		break;

	case 0x0D:  /* <CR> - just blanks out command window */
		break;

	case 0x17:  /* '^W' - setting continuous report command */
		DisplayMenuChoices (REPEAT_MENU, &kb_cmd);
		if (kb_cmd.kbch != 0x1B)
      {
			repeat_report_code = kb_cmd.replycode;
	      if (repeat_report_code == 0x120)
   	   {
      		cmd_0x35s (&cmd, 0x20, 0, 0, 0);
            send_cmd (&cmd, sendb);
         }
		}
		break;

	case '?':   /* '?' - Keystroke Command List */
		DisplayMenuChoices (ALL_MENU, &kb_cmd);
		if (kb_cmd.kbch != 0x1B) {
			switch_window (AUTOWIN);
			ProcessChatCommand (kb_cmd.kbch);		/* recursive call */
		}
		break;

	case '!':
		/* Creates a file GPSALM.DAT that contains all information */
		/* required to do a warm start. */
		cprintf ("Transfer GPSALM.DAT from rcvr to PC?");
		clreol ();
		if (!ask_verify()) break;
		if (!almgetb ()) {
			show_crlf();
			cprintf ("Storage aborted");
		}
		break;

	case '@':
		/* Load a file GPSALM.DAT that contains all information */
		/* required to do a warm start. */
		cprintf ("Transfer GPSALM.DAT from PC to rcvr?");
		clreol ();
		if (!ask_verify ()) break;
		if (!almputb ()) {
			show_crlf();
			cprintf ("Loading aborted");
		}
	   break;

	case '~':
		/* send a stinger break sequence */
		cprintf ("Send Break Sequence to rcvr?");
		clreol ();
		if (!ask_verify ()) break;
      pdef.baud = 9600;
      pdef.parity = 'n';
      pdef.databits = 8;
      pdef.stopbits = 1;
		ChangeSIOParms (&pdef);
		Do7xxBreak (300);
	   break;

	case 0x1B:
		/* open up command window */
		highvideo ();
		show_crlf();
		show_crlf();
		cprintf (
"******** Press ESC to exit program; or press any other key to continue *********");
		show_crlf();
		show_crlf();
		normvideo ();
		break;

	case '%':	/* unsigned char string */
		{
			unsigned char cmdbuflen, cmdbuf[21];

			cmdbuflen = ask_hex_string (
         	"TSIP command string (ID, then 0-19 bytes): ",
         	cmdbuf, 21);
         if (!cmd_esc() && cmdbuflen > 0)
         {
				cmd_0xxx (&cmd, cmdbuf[0], &cmdbuf[1], (unsigned char)(cmdbuflen-1));
   	   	send_cmd (&cmd, sendb);
         }
		}
		break;

	case '|':	/* unsigned char string */
		{
			int cmdbuflen, i;
			char cmdbuf[60];

			cmdbuflen = ask_char_string ("command characters: ", cmdbuf);
			for (i = 0; i < cmdbuflen; i++) sendb (cmdbuf[i]);
		}
		break;

	default: /* normal TSIP commands */
		ProcessTSIPKeystroke (&cmd, kbch, reply);
		if (cmd.code)
      {
      	send_cmd (&cmd, sendb);
      }
		break;
	}
	/* return to auto-report mode */
	switch_window (AUTOWIN);
}

static int ParseCommandLine (int argc, char *argv[])
{
	int
		iarg, retval;
	char
		argflag, *argptr;
   int
   	portnum;

	retval = TRUE;
	/* provisions for command line arguments starting with '-' or '/' */
	for (iarg = 0; iarg < argc; iarg++) {
		argflag = argv[iarg][0];
		if (argflag != '-' && argflag != '/') continue;
      argptr = &argv[iarg][2];
      switch (argv[iarg][1])
      {
      case 'c': case 'C':
			portnum = *argptr - '0';
			/* returns TRUE if serial port failure */
         if (portnum < 0 || portnum > 2) return TRUE;
			pdef.port = portnum;
	      pdef.baud = 9600;
   	   pdef.databits = 8;
      	pdef.parity = 'o';
	      pdef.stopbits = 1;

			if (initsio (portnum))
         {
         	printf ("\nInitial open port failed");
            closeserial();
            exit(0);
         }
         if (ChangeSIOParms (&pdef))
         {
         	printf ("\nInitial set port failed");
            closeserial();
            exit(0);
         }
         retval = FALSE;
         break;
      case 'f': case 'F':
			strcpy (tsipfilename, argptr);
			open_file_storage (&tsipstream, tsipfilename);
      }
	}
   if (retval)
   {
		show_crlf();show_crlf();
		printf ("\nSyntax error. Correct Syntax:");
		printf ("\nTSIPCHAT -c<n> -f<filename>");
		printf ("\nComm Port must be specified by command line argument");
		printf ("\n-c1 for COM1, etc.");
		printf ("\nBinary output file name optional: -f<filename>");
   }
	return retval;
}

/**/
/* ************************************************************** */
/* ****************  main routine for tsiptalk ****************** */
/* ************************************************************** */

void main (int argc, char *argv[])
{
	int
   	i,
      last_kbch,
      kbch;

	new_repeat_report = FALSE;
   tsipchat = TRUE;
	repeat_report_code = 0;
	nprintchar = 0;
	for (i=0; i<0x200; i++) reply[i] =0;
	serial_port_hold = FALSE;
	tsipstream = NULL;

	if (ParseCommandLine(argc, argv))
	{
		exit(0);
	}

   /* make sure serial ports are closed nicely at exit */
   atexit(closeserial);

	/* clear screen and start at the bottom */
	initialize_screen ();
   atexit(reset_screen);

	/* initialize two screen setup */
	switch_window (CMDWIN);
	highvideo ();
	cprintf (
		"       TSIPCHAT %s(%s), TSIP Receiver Control Interface Software",
		TSIPCHAT_VERNUM, TSIP_VERNUM);
	show_crlf ();
	cprintf (
		"              (c) Copyright 1993-9 Trimble Navigation, Ltd.");
	show_crlf ();
	cprintf (
		"              Press '?' to access the Keystroke Command List");
	show_crlf ();
	cprintf (
		"                  Press ESC twice to exit the program");
	normvideo ();

	switch_window (AUTOWIN);

	signal (SIGINT, ctrlchandler);

	last_kbch = 0;

	/* MAIN LOOP:  read bytes from buffered serial port input; */
	/* if kbhit, do corresponding command */
	/* double ESC to quit program */

	for (;;)
	{
		kbch = ProcessInputBytesWhileWaitingForKBHit ();
		if (kbch==0x1B && last_kbch==0x1B) break;
		ProcessChatCommand (kbch);
		last_kbch = kbch;
   }
}

