/*
 * *************************************************************************
 *
 * Trimble Navigation, Ltd.
 * OEM Products Development Group
 * P.O. Box 3642
 * 645 North Mary Avenue
 * Sunnyvale, California 94088-3642
 *
 * Corporate Headquarter:
 *    Telephone:  (408) 481-8000
 *    Fax:        (408) 481-6005
 *
 * Technical Support Center:
 *    Telephone:  (800) 767-4822	(U.S. and Canada)
 *                (408) 481-6940    (outside U.S. and Canada)
 *    Fax:        (408) 481-6020
 *    BBS:        (408) 481-7800
 *    e-mail:     trimble_support@trimble.com
 *		ftp://ftp.trimble.com/pub/sct/embedded/bin
 *
 *
 * TSIP_CMD.C carries the function help_screen() displays the Keystroke
 * Command List in the format (key packet Id description).  For example,
 * (t 0x21 show time) means the keystroke 't' must be pressed in order to
 * request the time the receiver carries whether it is from an upload from
 * the user/computer or from the GPS constellation.  The TSIP commands are
 * displayed in the Keystroke Command List in ascending hexadecimal order.
 *
 * proc_kbd() parses the keyboard inputs and
 * directs processing to the mid-level function specific to the keyboard
 * stroke.  The called function then queries for user parameters and passes
 * them to the appropriate command string generator cmd_0x?? in TSIP_IFC.C.
 * The suffix convention for the command string generator follows:
 *
 *                   cmd_0x??q   -  query
 * 	               cmd_0x??s   -  set
 * 	               cmd_0x??c   -  clear/set to default
 * 	               cmd_0x??e   -  enable
 * 	               cmd_0x??d   -  disable
 *
 * In general, one command code corresponds to one keystroke.
 * *************************************************************************
 *
 */

#include <stdio.h>
#include <conio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include "t_struct.h"
#include "t_format.h"
#include "tsipincl.h"
#include "t_labels.h"

extern short tsipchat;
static char
	*databit_text_app [] =
		{"", "", "seven data bits", "eight data bits"},
	*stopbit_text[] =
		{"one stopbit ", "two stopbits"},
	*req_set_text[] =
		{"Request", "Set    "},
	*toggle_text[] =
		{"off", "on "};
/*********************************************************************************************/
/*********************************************************************************************/

#define MAX_KB_CMDS 100

/* Generate Control-letter character */
#define CNTRLKEY ('A'-0x01)
#define cntrl(x) ((x)-CNTRLKEY)
static int
	  num_replies;

/* Utility to check on sv_prn values */
static int prn_valid (unsigned char sv_prn)
{
	if ((sv_prn > 0) && (sv_prn < 33)) {
	  return TRUE;
   }
	cprintf ("SV number must be between 1 and 32 inclusive");
	return FALSE;
}

/**/
/* ***************  USER-PROMPTING ROUTINES FOR COMMANDS ************** */

/* 0x1D */
static void set_osc_offset (TSIPPKT *pCmd)
{
	float
		offset;
	unsigned char
		op_code;
	static char
		*op_text[2] = {"clear", "set"};

	op_code = PickFromMenu ("Clear or set oscillator: ", op_text, sizeof(op_text));
	if (cmd_esc ()) return;

	switch (op_code)
	{
	case 0:
		cmd_0x1Dc (pCmd);
		break;

   case 1:
		offset = ask_dbl ("Set osc offset to (Hz, relative to 1575.42 MHz): ");
		if (cmd_esc ()) return;
		cmd_0x1Ds (pCmd, offset);
   	break;
   }
}

/* 0x1E */
static void cold_start (TSIPPKT *pCmd)
{
	static char *reset_text [3] = {
		"Erase RAM, Restart",
		"Erase RAM, Set SEEROM to Factory Settings, Restart" ,
		"Erase RAM, Set SEEROM to v7.02 Settings, Restart" };
	short			iResetCode ;

   iResetCode = PickFromMenu ("Enter Reset Type: ", reset_text, sizeof(reset_text));
   if (cmd_esc ()) return;
   switch (iResetCode)
   {
   case 0: cmd_0x1E (pCmd, 0x4B); break;
   case 1: cmd_0x1E (pCmd, 0x46); break;
   case 2: cmd_0x1E (pCmd, 0x43); break;
   }
}
/**/
/* 0x20 */
static void request_almanac (TSIPPKT *pCmd)
{
	unsigned char
		sv_prn;

	sv_prn = ask_byte ("Enter SV PRN number: ");
   if (!prn_valid (sv_prn)) {
   	set_cmd_esc (TRUE);
   }
	if (cmd_esc ()) return;
	cmd_0x20 (pCmd, sv_prn);
	return;
}

/* 0x22 */
static void set_nav_mode (TSIPPKT *pCmd)
{
	unsigned char
		NavMode;
   static char
		*NavModeText0x22[] =	{
      	"automatic",
         "Time Only (One SV)", "",
         "2-D",
			"3-D", "", "", "", "", "",
         "Time Only (Multi-SV Avg)"};

	NavMode = PickFromMenu ("Enter nav mode: ", NavModeText0x22, sizeof(NavModeText0x22));
	if (cmd_esc ()) return;
	cmd_0x22 (pCmd, NavMode);
}

/* 0x23 */
static void set_initial_ECEF_position (TSIPPKT *pCmd)
{
	float
		pos_ECEF[3];

	pos_ECEF[0] = ask_dbl ("Enter ECEF x: ");	if (cmd_esc ()) return;
	pos_ECEF[1] = ask_dbl ("Enter ECEF y: ");	if (cmd_esc ()) return;
	pos_ECEF[2] = ask_dbl ("Enter ECEF z: ");	if (cmd_esc ()) return;

	cmd_0x23 (pCmd, pos_ECEF);
	return;
}

/* 0x2A */
static void set_altitude(TSIPPKT *pCmd)
{
	float
		alt;
	unsigned char
		op_code;
	static char
		*op_text[3] = {"", "set alt", "disable"};

	op_code = PickFromMenu ("Enter operation: ", op_text, sizeof(op_text));

	if (cmd_esc ()) return;

	switch (op_code)
	{
	case 1:
		alt = ask_dbl("Enter reference altitude: ");
		if (cmd_esc ()) break;
		cmd_0x2As (pCmd, alt) ;
		break;
	case 2:
		cmd_0x2Ad (pCmd);
		break;
	}

	return ;
}

/* 0x2B */
static void set_initial_lla_position (TSIPPKT *pCmd)
{
	float
		lat, lon, alt;

	lat = ask_dbl ("Enter lat (, negative for south): ");
	if (cmd_esc ()) return;
	lon = ask_dbl ("Enter lon (, negative for west): ");
	if (cmd_esc ()) return;
	alt = ask_dbl ("Enter alt (m): ");
	if (cmd_esc ()) return;

	lat *= D2R;
	lon *= D2R;
	cmd_0x2B (pCmd, lat, lon, alt);
	return;
}

static void set_nav_params (TSIPPKT *pCmd)
{
	unsigned char
		dyn_code, set_req;
	float
		elev_mask, snr, dop_mask, dop_switch;

	set_req = PickFromMenu ("Navigation Mode Parameters: ", req_set_text, sizeof(req_set_text));
	if (cmd_esc ()) return;
	if (!set_req)
	{
		cmd_0x2Cq (pCmd);
      return;
   }

	dyn_code = PickFromMenu ("Enter dynamics code: ", dyn_text, sizeof(dyn_text));
	if (cmd_esc ()) return;
	elev_mask = ask_dbl ( "Enter elev angle mask, -1.0 for no change: ");
	if (cmd_esc ()) return;
	snr = ask_dbl (       "Enter SNR mask       , -1.0 for no change: ");
	if (cmd_esc ()) return;
	dop_mask = ask_dbl (  "Enter DOP mask       , -1.0 for no change: ");
	if (cmd_esc ()) return;
	dop_switch = ask_dbl ("Enter PDOP switch    , -1.0 for no change: ");
	/* Convert elevation mask to radians */
	elev_mask *= D2R;
	cmd_0x2Cs (pCmd, dyn_code, elev_mask, snr, dop_mask, dop_switch);
	return;
}
/* 0x2E */
static void set_time (TSIPPKT *pCmd)
{
	short
		week_num;
	double
		gpstime;
	float
		time_of_week;
	short	days, hours, minutes;
	float seconds;


	/* set time from PC */
	GPS_time_from_PC (&gpstime);
	if (cmd_esc ()) return;
	week_num = (short) (gpstime/604800.);
	time_of_week = (gpstime - 604800.*week_num);
	cprintf ("GPS week %3d", week_num);
	days = (short)(time_of_week / 86400.0);
	hours = (short)fmod(time_of_week / 3600., 24.);
	minutes =  (short) fmod(time_of_week/60., 60.);
	seconds = fmod(time_of_week, 60.);
	cprintf (" %s %02d:%02d:%05.2f   ", dayname[days], hours, minutes, seconds);
	show_crlf ();
	cmd_0x2E (pCmd, time_of_week, week_num);
	return;
}


/**/
/* 0x31 */
static void set_accurate_ECEF_position (TSIPPKT *pCmd)
{
	float
		pos_ECEF[3];

	cprintf ("CAUTION: Position should be accurate");
	show_crlf ();
	pos_ECEF[0] = ask_dbl ("Enter ECEF x: ");
	if (cmd_esc ()) return;
	pos_ECEF[1] = ask_dbl ("Enter ECEF y: ");
	if (cmd_esc ()) return;
	pos_ECEF[2] = ask_dbl ("Enter ECEF z: ");
	if (cmd_esc ()) return;
	cmd_0x31 (pCmd, pos_ECEF);
	return;
}

/* 0x32 */
static void set_accurate_LLA_position (TSIPPKT *pCmd)
{
	float
		lat, lon, alt;

	cprintf ("CAUTION: Position should be accurate");
	show_crlf ();
	lat = ask_dbl ("Enter lat (, negative for south): ");
	if (cmd_esc ()) return;
	lon = ask_dbl ("Enter lon (, negative for west): ");
	if (cmd_esc ()) return;
	alt = ask_dbl ("Enter alt (m): ");
	if (cmd_esc ()) return;
	/* Convert lat and lon to radians */
	lat *= D2R;
	lon *= D2R;
	cmd_0x32 (pCmd, lat, lon, alt);
	return;
}

/* 0x34 */
static void SetSVForOneSVMode (TSIPPKT *pCmd)
{
	unsigned char
		sv_prn;

	sv_prn = ask_byte ("Enter SV PRN number (0 = always pick highest): ");
   if (sv_prn && !prn_valid (sv_prn)) {
   	set_cmd_esc (TRUE);
   }
	if (cmd_esc ()) return;
	cmd_0x34 (pCmd, sv_prn);
	return;
}

/* 0x35 */
static void set_IO_options (TSIPPKT *pCmd)
{
	unsigned char
		hexnum, hexbyte[5];
	static char
		*no_yes_text[] = {"no", "yes"},
		*alt_text[] = {"WGS-84", "MSL   "},
		*prec_text[] = {"single", "double"},
		*time_text[] = {"GPS", "UTC"},
      *query_text[] = {"auto-report", "poll only"},
		*raw_text[] = {"raw (unfiltered)", "Doppler-smoothed"},
		*signal_text[] = {"AMU", "dBHz"};

	short
   	bit0, bit1, bit2, bit3, bit4, bit5,
		pos_code, vel_code, time_code, aux_code;

	unsigned char
   	req_set;

	req_set = PickFromMenu("Query or Set?  ", req_set_text, sizeof(req_set_text));
	if (cmd_esc()) return;
	if (req_set == 0)
	{
		cmd_0x35q(pCmd) ;
		return;
	}

	if (!PickFromMenu("Enter full menu?", no_yes_text, sizeof(no_yes_text)))
	{
		if (cmd_esc ()) return;
		hexnum = ask_hex_string ("Default is 1 1 0 0: ", hexbyte, 4);
		if (hexnum != 4)
		{
			set_cmd_esc (TRUE);
		}
		else
		{
			pos_code = hexbyte[0];
			vel_code = hexbyte[1];
			time_code = hexbyte[2];
			aux_code = hexbyte[3];
		}
		if (cmd_esc ()) return;
	}
	else
	{ /* position code */
		bit0 = PickFromMenu ("ECEF position output: ", toggle_text, sizeof(toggle_text));
		if (cmd_esc ()) return;
		bit1 = PickFromMenu ("LLA position output: ", toggle_text, sizeof(toggle_text));
		if (cmd_esc ()) return;
		if (bit1 == 1)
		{
			bit2 = PickFromMenu ("altitude output: ", alt_text, sizeof(alt_text));
			if (cmd_esc ()) return;
		}
		else
		{
			bit2 = 0;
		}
		bit3 = PickFromMenu ("altitude input: ", alt_text, sizeof(alt_text));
		if (cmd_esc ()) return;
		bit4 = PickFromMenu ("output precision: ", prec_text, sizeof(prec_text));
		if (cmd_esc ()) return;
		bit5 = PickFromMenu ("Superpacket output: ", toggle_text, sizeof(toggle_text));
		if (cmd_esc ()) return;
		pos_code = (unsigned char)
      	(bit0+(2*bit1)+(4*bit2)+(8*bit3)+(16*bit4)+(32*bit5));

		/* velocity code */
		bit0  = PickFromMenu ("ECEF velocity output: ", toggle_text, sizeof(toggle_text));
		if (cmd_esc ()) return;
		bit1  = PickFromMenu ("ENU velocity output: ", toggle_text, sizeof(toggle_text));
		if (cmd_esc ()) return;
		vel_code = (unsigned char)(bit0 + (2 * bit1));

		/* time code */
		bit1 = bit3 = bit4 = 0;
		bit0 = PickFromMenu ("Time tags in: ", time_text, sizeof(time_text));
		if (cmd_esc ()) return;
		bit2 = PickFromMenu ("Output time: ", query_text, sizeof(query_text));
		if (cmd_esc ()) return;
		time_code = (unsigned char)
      	(bit0 + (2 * bit1) + (4 * bit2) + (8 * bit3) + (16 * bit4));

		/* auxiliary code */
		bit1 = bit2 = bit4 = 0;
		bit0 = PickFromMenu ("Pseudorange output: ", toggle_text, sizeof(toggle_text));
		if (cmd_esc ()) return;
		if (bit0 == 1) {
			bit1  = PickFromMenu ("Pseudorange quality: ", raw_text, sizeof(raw_text));
			if (cmd_esc ()) return;
		}
		bit3 = PickFromMenu ("Signal strength output units: ", signal_text, sizeof(signal_text));
		if (cmd_esc ()) return;
		aux_code = (unsigned char)
      	(bit0 + (2 * bit1) + (4 * bit2) + (8 * bit3));
	}
	cmd_0x35s (pCmd, pos_code, vel_code, time_code, aux_code);
}

/* 0x38 */
static void request_SV_system_data (TSIPPKT *pCmd)
{
	unsigned char
		type_of_data,
		sv_prn;
	static char
		*type_text[7] = {
      	"",
         "",
         "almanac",
         "health page, t_oa, wn_oa",
         "ionosphere",
         "UTC",
         "ephemeris"};

   /*
	* For loading, a data buffer must be supplied from somewhere
	* with the information to be loaded.  Because misloading of
	* data can cause the receiver to act erratic, the load
	* feature is not implemented here.
	* cmd_0x38s (pCmd, type_of_data, sv_prn, nbytes, databuf);
	*/
	type_of_data = PickFromMenu ("Type of data: ", type_text, sizeof(type_text));
	if (cmd_esc ()) return;

	if ((type_of_data == 2) || (type_of_data == 6)) {
		sv_prn = ask_byte ("Enter SV number: ");
		if (!prn_valid(sv_prn)) set_cmd_esc (TRUE);
		if (cmd_esc ()) return;
	} else {
	  sv_prn = 0;
   }
	cmd_0x38q (pCmd, type_of_data, sv_prn);
}


/* 0x39 */
static void request_disable_or_health_override_list (TSIPPKT *pCmd)
{
	unsigned char
		op_code, sv_prn;
	static char
		*op_text[] = {"",
      	"enable SV",
         "disable SV",
			"list enabled/disabled",
         "heed SV health",
			"ignore SV health",
         "list heed/ignore health"};

	op_code = PickFromMenu ("Enter operation code: ", op_text, sizeof(op_text));
	if (cmd_esc ()) return;

	sv_prn = 0;
	if ((op_code != 3) && (op_code != 6)) {
		sv_prn = ask_byte ("Enter SV PRN number or 0 for all:  ");
      if (sv_prn && !prn_valid(sv_prn))
      {
			set_cmd_esc (TRUE);
		}
		if (cmd_esc ()) return;
      num_replies = 0;
	}
	cmd_0x39 (pCmd, op_code, sv_prn);
	return;
}

/* 0x3A */
static void request_raw_msmt_data (TSIPPKT *pCmd)
{
	unsigned char
		sv_prn;

	sv_prn = ask_byte ("Enter SV PRN number or 0 for all: ");
	if (sv_prn && !prn_valid (sv_prn)) set_cmd_esc (TRUE);
	if (cmd_esc ()) return;
	cmd_0x3A (pCmd, sv_prn);
	num_replies = (sv_prn)?1:8;
	return;
}

/* 0x3B */
static void request_eph_status (TSIPPKT *pCmd)
{
	unsigned char
		sv_prn;

	sv_prn = ask_byte ("Enter SV PRN number or 0 for all: ");
	if (sv_prn && !prn_valid (sv_prn)) set_cmd_esc (TRUE);
	if (cmd_esc ()) return;
	cmd_0x3B (pCmd, sv_prn);
	num_replies = (sv_prn)?1:32;
	return;
}

/* 0x3C */
static void request_track_info (TSIPPKT *pCmd)
{
	unsigned char
		sv_prn;

	sv_prn = ask_byte ("Enter SV PRN number or 0 for all: ");
	if (sv_prn && !prn_valid (sv_prn)) set_cmd_esc (TRUE);
	if (cmd_esc ()) return;
	cmd_0x3C (pCmd, sv_prn);
	num_replies = (sv_prn)?1:8;
	return;
}

/* 0x3D */
static void old_set_port_config (
   unsigned char *baud_out,
   unsigned char *baud_inp,
   unsigned char *char_code,
	unsigned char *stopbitcode,
   unsigned char *output_mode,
   unsigned char *input_mode)
{
	int
		i, j;
	static char
		*stop_bit_text[]       =  {"", "one", "two"};

	i = PickFromMenu ("Enter XMT Baud Rate: ", st_baud_text_app, sizeof(st_baud_text_app));
	if (cmd_esc ()) return;
	*baud_out = old_baudnum[i];

	i = PickFromMenu ("Enter RCV Baud Rate: ", st_baud_text_app, sizeof(st_baud_text_app));
	if (cmd_esc ()) return;
	*baud_inp = old_baudnum[i];

	i = PickFromMenu ("Enter bits/char:     ", databit_text_app, sizeof(databit_text_app));
	if (cmd_esc ()) return;

	j = PickFromMenu ("Enter parity:        ", old_parity_text, sizeof(old_parity_text));
	if (cmd_esc ()) return;
	*char_code = (unsigned char)(i + (j << 2));

	i = PickFromMenu ("Enter stopbits:      ", stop_bit_text, sizeof(stop_bit_text));
	if (cmd_esc ()) return;
	if (i == 1)             /* 0x07 */
		*stopbitcode = 0x07;
	else if (i == 2)        /* 2 stop bits code */
		*stopbitcode = 0x0F;

	i = PickFromMenu ("Enter XMT language:  ", old_output_ch, sizeof(old_output_ch));
	if (cmd_esc ()) return;
	*output_mode = (short)i;
	i = PickFromMenu ("Enter RCV language:  ", old_input_ch, sizeof(old_input_ch));
	if (cmd_esc ()) return;
	*input_mode = (short)i;
}

static void set_chan_A (TSIPPKT *pCmd)
{
   unsigned char
   	set_req;
   unsigned char
   	baud_out,
   	baud_inp,
   	char_code,
		stopbitcode,
   	output_mode,
   	input_mode;

	set_req = PickFromMenu ("Port A configuration: ", req_set_text, sizeof(req_set_text));
	if (cmd_esc ()) return;
	if (!set_req)
	{
		cmd_0x3Dq (pCmd);
      return;
   }

	old_set_port_config (&baud_out, &baud_inp, &char_code,
   	&stopbitcode, &output_mode, &input_mode);
   cmd_0x3Ds (pCmd, baud_out, baud_inp, char_code,
      stopbitcode, output_mode, input_mode);
}
/* 0xBB */

static void set_stinger_rcvr_config (TSIPPKT *pCmd)
{
	TSIP_RCVR_CFG
   	TsipBB;

	unsigned char req_set;

   TsipBB.bSubcode = 0;

	req_set = PickFromMenu("Receiver Configuration:  ", req_set_text, sizeof(req_set_text));
	if (cmd_esc()) return;
	if (req_set == 0)
	{
		cmd_0xBBq(pCmd, TsipBB.bSubcode) ;
		return;
	}

	TsipBB.operating_mode = PickFromMenu ("Operating mode:  ", NavModeText0xBB, sizeof(NavModeText0xBB));
	if (cmd_esc()) return;
	TsipBB.dyn_code = PickFromMenu ("Dynamics code:   ", dyn_text, sizeof(dyn_text));
	if (cmd_esc()) return;
	TsipBB.elev_mask = ask_dbl ("Enter elev angle mask (deg, -1 = no change): ");
	TsipBB.elev_mask *= D2R;	/* Convert elevation mask to radians. */
	if (cmd_esc()) return;
	TsipBB.cno_mask = ask_dbl ("Enter SNR mask (AMUs, -1 = no change):   ");
	if (cmd_esc()) return;
	TsipBB.dop_mask = ask_dbl ("Enter DOP mask (-1 = no change):       ");
	if (cmd_esc()) return;
	TsipBB.dop_switch = ask_dbl ("Enter DOP switch (-1 = no change):     ");
	if (cmd_esc()) return;

	cmd_Paly0xBBs(pCmd, &TsipBB) ;
}


/* 0xBC */
static void set_rcvr_serial_port_config (TSIPPKT *pCmd)
{
	unsigned char
		port_num,
		in_baud,
		out_baud,
		data_bits,
		parity,
		stop_bits,
		flow_control,
		protocols_in,
		protocols_out,
		reserved,
		temp,
		req_set;

	req_set = PickFromMenu("Port Configuration:  ", req_set_text, sizeof(req_set_text));
	if (cmd_esc()) return;
	port_num = PickFromMenu ("Port: ", rcvr_port_text, sizeof(rcvr_port_text));
	if (port_num != 0 && port_num != 1) port_num = 0xFF;
	if (cmd_esc()) return;

	if (req_set == 0)
	{
		cmd_0xBCq (pCmd, port_num);
		return;
	}
	/* To set the "current port", use a port number of 0xFF */
	cprintf ("Set RECEIVER Port Configuration");
   show_crlf();
   if (port_num == 0)
   {
		temp = PickFromMenu      ("Protocol in:      ",
      	protocols_in_text, sizeof(protocols_in_text));
		if (cmd_esc()) return;
		protocols_in = (unsigned char)(1 << temp);
		temp = PickFromMenu      ("Protocol out:     ",
      	protocols_out_text, sizeof(protocols_out_text));
		if (cmd_esc()) return;
		protocols_out = (unsigned char)(1 << temp);
   }
   else
   {
		cprintf ("Protocol in:      TSIP"); show_crlf();
   	protocols_in = 0x02;
		cprintf ("Protocol out:     TSIP"); show_crlf();
   	protocols_out = 0x02;
   }
	in_baud = PickFromMenu      ("Input baud rate:  ",
   	st_baud_text_app, sizeof(st_baud_text_app));
	if (cmd_esc()) return;
	out_baud = PickFromMenu     ("Output baud rate: ",
   	st_baud_text_app, sizeof(st_baud_text_app));
	if (cmd_esc()) return;
   /* receiver will freeze up if set to 7 bits and TSIP */
   if ((protocols_in & 0x02) || (protocols_out & 0x02))
   {
		cprintf ("Data bits:         eight"); show_crlf();
		data_bits = 3;
   }
   else
   {
		data_bits = PickFromMenu ("Data bits:        ",
      	databit_text_app, sizeof(databit_text_app));
	}
	if (cmd_esc()) return;
	parity = PickFromMenu       ("Parity:           ",
   	parity_text, sizeof(parity_text));
	if (cmd_esc()) return;
	stop_bits = PickFromMenu    ("Stop bits:        ",
   	stopbit_text, sizeof(stopbit_text));
	if (cmd_esc()) return;
	flow_control = 0;
	show_crlf();
   reserved = 0;

   cprintf ("Warning! Record New Serial Port settings.  Continue?");
	if (!ask_verify ()) {
		set_cmd_esc (TRUE);
		return;
	}

	cmd_0xBCs (pCmd, port_num, in_baud, out_baud, data_bits, parity, stop_bits,
		flow_control, protocols_in, protocols_out, reserved);
}


/**/
static void Control8F0B (TSIPPKT *pCmd)
{
	int
   	set_req,
      EventOut,
      PPSOut;
   unsigned char
   	mask;


	set_req = PickFromMenu ("Port A 0x8F-0B Output Control: ", req_set_text, sizeof(req_set_text));
	if (cmd_esc ()) return;
	if (!set_req)
	{
	  	cmd_0x8E0Bq (pCmd);
      return;
   }

   PPSOut   = PickFromMenu ("PPS Output  : ",
   	toggle_text, sizeof(toggle_text));
   if (cmd_esc()) return;
   EventOut = PickFromMenu ("Event Output: ",
   	toggle_text, sizeof(toggle_text));
   if (cmd_esc()) return;

   mask = (unsigned char)((EventOut<<1) | PPSOut);

  	cmd_0x8E0Bs (pCmd, mask);
   num_replies = 0;

}
/* 0x8E-14 */
static void set_datum (TSIPPKT *pCmd)
{
	short datum_index;

	datum_index = (short)ask_int ("Enter Desired Datum index (-1 for custom) :  ");
	if (cmd_esc ()) return;
	if (datum_index >= 0 && datum_index < 256)
	{
		cmd_0x8E14s (pCmd, datum_index);
		return;
	}
	else
	{
		double datum_coeffs[5];
		/* Get 5 double values */
      cprintf ("Input custom datum parameters:");
      show_crlf();
		datum_coeffs[0] = ask_dbl("Enter value for DX        :");
		datum_coeffs[1] = ask_dbl("Enter value for DY        :");
		datum_coeffs[2] = ask_dbl("Enter value for DZ        :");
		datum_coeffs[3] = ask_dbl("Enter value for A-axis    :");
		datum_coeffs[4] = ask_dbl("Enter value for E-squared :");
		cmd_0x8E14s_2 (pCmd, datum_coeffs);
	}
}

/* 0x8E-20 */
static void superpacket_20_control (TSIPPKT *pCmd)
{
	unsigned char
		mode, cmdbuflen, cmdbuf[3];
	static char
		*superpacket_enable[3] = {"off  ", "on   ", "query"};


	mode = PickFromMenu ("Superpacket output:  ", superpacket_enable, sizeof(superpacket_enable));
	if (cmd_esc ()) return;
	if (mode > 1) {
		cmdbuflen = 1;
		cmdbuf[0] = 0x20;
	}
   else
   {
		cmdbuflen = 2;
		cmdbuf[0] = 0x20;
		cmdbuf[1] = mode;
      num_replies = 0;
	}
	cmd_0x8E (pCmd, cmdbuf, cmdbuflen);
}

static void RevertToDefaults (TSIPPKT *pCmd)
{
	unsigned char
   	segment;

	static char
		*OptionsPrefixText[] = {"", "", "",
      	"CNFG", "", "PORT", "PPS",
         "ACCU", "DECORR", "TIMING"};


	segment = PickFromMenu ("Memory Segment to clear:  ",
   	OptionsPrefixText, sizeof(OptionsPrefixText));
	if (cmd_esc ()) return;
   cmd_0x8E45 (pCmd, segment);
}

static void SetPPSParms (TSIPPKT *pCmd)
{
	int
   	set_req;
   unsigned char
   	OnOff,
   	TimeBase,
      Polarity;
   double
   	Offset;
   float
   	Uncertainty;

	set_req = PickFromMenu ("PPS Parameters: ", req_set_text, sizeof(req_set_text));
	if (cmd_esc ()) return;
	if (!set_req)
	{
	  	cmd_0x8E4Aq (pCmd);
      return;
   }

   OnOff = PickFromMenu ("On or Off?  ", toggle_text, sizeof(toggle_text));
   if (cmd_esc()) return;
   TimeBase = PickFromMenu ("Time Base:  ", PPSTimeBaseText, sizeof(PPSTimeBaseText));
   if (cmd_esc()) return;
   Polarity = PickFromMenu ("Polarity:  ", PPSPolarityText, sizeof(PPSPolarityText));
   if (cmd_esc()) return;
   Offset = ask_dbl ("PPS Offset or Cable Delay (nsecs):  ");
   if (cmd_esc()) return;
   Offset *= 1.0e-9;
   Uncertainty = ask_dbl ("Max Allowable PPS Uncertainty (nsecs):   ");
   if (cmd_esc()) return;
   Uncertainty *= GPS_C * 1.e-9;

   cmd_0x8E4As (pCmd, OnOff, TimeBase, Polarity, Offset, Uncertainty);
   num_replies = 0;
}

static void SurveyLimit (TSIPPKT *pCmd)
{
	int
   	set_req;
   long
   	AutoSurveyLimit;

	set_req = PickFromMenu ("AutoSurvey Parameter: ", req_set_text, sizeof(req_set_text));
	if (cmd_esc ()) return;
	if (!set_req)
	{
	  	cmd_0x8E4Bq (pCmd);
      return;
   }

   AutoSurveyLimit = ask_lint ("Max # of position fixes for self-survey : ");
   if (cmd_esc()) return;
   cmd_0x8E4Bs (pCmd, AutoSurveyLimit);
}

static void PortBOutputMask (TSIPPKT *pCmd)
{
	int
   	ichoice, numchoices,
   	numbytes,
   	bit,
   	set_req;
   static char
		*no_yes_text[] = {"no", "yes"};
   unsigned long
   	OutputMask;
   unsigned char OMBytes[4];
   static unsigned long
   	MaskBit[] = {
      	0x00000001, 0x00000002, 0x00000004, 0x00000008, 0x00000010, 0x00000020,
      	0x00000100L, 0x00000800L, 0x00001000L,
         0x40000000L, 0x80000000L};

	set_req = PickFromMenu ("Port B Output Mask: ", req_set_text, sizeof(req_set_text));
	if (cmd_esc ()) return;
	if (!set_req)
	{
	  	cmd_0x8E4Dq (pCmd);
      return;
   }

	if (!PickFromMenu("Enter full menu?  ", no_yes_text, sizeof(no_yes_text)))
	{
		if (cmd_esc ()) return;
		do {
      	numbytes = ask_hex_string ("Default is FF FF FF C0: ", OMBytes, 4);
			if (cmd_esc ()) return;
      } while (numbytes != 4);
      OutputMask = ((unsigned long)OMBytes[0]<<24) | ((unsigned long)OMBytes[1]<<16) |
      	((unsigned long)OMBytes[2]<<8) | ((unsigned long)OMBytes[3]);
	}
	else
	{
   	/* toggle reserved bits on */
      OutputMask = 0x3FFFE6C0L;
      numchoices = sizeof(MaskBit)/sizeof(unsigned long);
      for (ichoice = 0; ichoice<numchoices; ichoice++)
      {
			bit = PickFromMenu (MaskText[ichoice], toggle_text, sizeof(toggle_text));
			if (cmd_esc ()) return;
      	if (bit) OutputMask |= MaskBit[ichoice];
      }
   }
   cmd_0x8E4Ds (pCmd, OutputMask);
}


static void Control8FAD (TSIPPKT *pCmd)
{
	int
   	set_req,
      EventOut,
      PPSOut;
   unsigned char
   	mask;


	set_req = PickFromMenu ("Port A 0x8F-AD Output Control: ", req_set_text, sizeof(req_set_text));
	if (cmd_esc ()) return;
	if (!set_req)
	{
	  	cmd_0x8EADq (pCmd);
      return;
   }

   PPSOut   = PickFromMenu ("Port A 8F-AD PPS Output:   ",
   	toggle_text, sizeof(toggle_text));
   if (cmd_esc()) return;
   EventOut = PickFromMenu ("Port A 8F-AD Event Output: ",
   	toggle_text, sizeof(toggle_text));
   if (cmd_esc()) return;

   mask = (unsigned char)((EventOut<<1) | PPSOut);

  	cmd_0x8EADs (pCmd, mask);
   num_replies = 0;

}
/**/

static int get_menu (int code, KB_CMD **kb_cmd_list)
{
static KB_CMD kb_all_cmd[] = {
	{0x17, 0x11, 0x00, "repeat command   "},
	{0x06, 0x11, 0x00, "log TSIP to file "},
	{ '!', 0x11, 0x00, "download almanac "},
	{ '@', 0x11, 0x00, "upload almanac   "},
	{ '~', 0x11, 0x00, "send serial break"},
	{0x09, 0x11, 0x00, "set PC serial prt"},
	{ '%', 0x10, 0x00, "TSIP cmd string  "},
	{ '=', 0x10, 0x00, "Superpkt 8Eh menu"},
	{ 'F', 0x1D, 0x00, "set osc offset   ", set_osc_offset},
	{0x0B, 0x1E, 0x45, "clear battery RAM", cold_start},
	{ 'v', 0x1F, 0x45, "show S/W version ", cmd_0x1F},
	{ 'A', 0x20, 0x40, "*show almanac    ", request_almanac},
	{ 't', 0x21, 0x41, "show time        ", cmd_0x21},
	{ 'N', 0x22, 0x00, "*set 2D/3D mode  ", set_nav_mode},
	{ 'i', 0x23, 0x00, "input XYZ pos    ", set_initial_ECEF_position},
	{ 'M', 0x24, 0x6D, "show 2D/3D mode  ", cmd_0x24},
	{0x12, 0x25, 0x45, "reset receiver   ", cmd_0x25},
	{ 'h', 0x26, 0x46, "receiver health  ", cmd_0x26},
	{ 's', 0x27, 0x47, "show sig levels  ", cmd_0x27},
	{ 'm', 0x28, 0x48, "GPS sys message  ", cmd_0x28},
	{ 'a', 0x29, 0x49, "*show alm health ", cmd_0x29},
	{0x01, 0x2A, 0x4A, "altitude for 2-D ", set_altitude},
	{ 'I', 0x2B, 0x00, "input LLA pos    ", set_initial_lla_position},
	{ 'P', 0x2C, 0x4C, "*set nav params  ", set_nav_params},
	{ 'o', 0x2D, 0x4D, "show osc offset  ", cmd_0x2D},
	{ 'T', 0x2E, 0x4E, "set time from PC ", set_time},
	{ 'u', 0x2F, 0x4F, "*show UTC data   ", cmd_0x2F},
	{ 'b', 0x31, 0x00, "set acc ECEF pos ", set_accurate_ECEF_position},
	{ 'B', 0x32, 0x00, "set acc LLA pos  ", set_accurate_LLA_position},
	{ 'S', 0x34, 0x00, "SV for 1-SV mode ", SetSVForOneSVMode},
	{ 'O', 0x35, 0x55, "set  I/O options ", set_IO_options},
	{ 'w', 0x37, 0x57, "show last fix    ", cmd_0x37},
	{ '$', 0x38, 0x58, "show SV data     ", request_SV_system_data},
	{ 'd', 0x39, 0x59, "enable/disable SV", request_disable_or_health_override_list},
	{ '<', 0x3A, 0x5A, "*show raw meas   ", request_raw_msmt_data},
	{ 'e', 0x3B, 0x5B, "*show eph status ", request_eph_status},
	{ '>', 0x3C, 0x5C, "show track status", request_track_info},
	{0x15, 0x3D, 0x3D, "*set port 1      ", set_chan_A},
	{ 'p', 0xBB, 0xBB, "rcvr configure   ", set_stinger_rcvr_config},
	{ 'U', 0xBC, 0xBC, "port configure   ", set_rcvr_serial_port_config},
	{'\0', 0x00, 0x00, "end record       "}
};

static KB_CMD kb_8e_cmd[] = {
	{ 'b', 0x0B, 0x100 + 0x0B, "8F-0B Control    ", Control8F0B},
	{ 'D', 0x14, 0x100 + 0x14, "set datum        ", set_datum},
	{ 'd', 0x15, 0x100 + 0x15, "query datum      ", cmd_0x8E15q},
	{ 'g', 0x20, 0x100 + 0x20, "Std PosVel spkt  ", superpacket_20_control},
	{ '1', 0x41, 0x100 + 0x41, "query Board SN   ", cmd_0x8E41q},
	{ '2', 0x42, 0x100 + 0x42, "query Product SN ", cmd_0x8E42q},
	{ 'k', 0x45, 0x100 + 0x45, "Revert to Deflts ", RevertToDefaults},
	{ 'p', 0x4A, 0x100 + 0x4A, "Set PPS parms    ", SetPPSParms},
	{ 'l', 0x4B, 0x100 + 0x4B, "Survey limit     ", SurveyLimit},
	{ 'o', 0x4D, 0x100 + 0x4D, "PortB Output Pkts", PortBOutputMask},
	{ 'm', 0xA5, 0x100 + 0xA5, "Spkt Output Mask ", cmd_0x8EA5q},
	{ 'a', 0xAD, 0x100 + 0xAD, "8F-AD Control    ", Control8FAD},
	{'\0', 0x00, 0x00, "end record       "}
};

static KB_CMD kb_repeat_cmd[] = {
	{ 't', 0x21, 0x41, "show time        "},
	{ 'o', 0x2D, 0x4D, "show osc offset  "},
	{ '<', 0x3A, 0x5A, "show raw meas    "},
	{ 's', 0x27, 0x47, "show sig levels  "},
	{ '>', 0x3C, 0x5C, "show track status"},
	{ '=',0x120,0x120, "Pos/Vel fixes    "},
	{'\0', 0x00, 0x00, "end record       "}
};

	switch (code) {
	case REPEAT_MENU:
		*kb_cmd_list = kb_repeat_cmd;
		break;
	case ALL_MENU:
		*kb_cmd_list = kb_all_cmd;
		break;
	case x8E_MENU:
		*kb_cmd_list = kb_8e_cmd;
		break;
	default:
		return 1;
	}
	return 0;
}

/* list of keyboard strokes and command each invokes */
/* When '?' is pressed, a command list is displayed and a FALSE is   */
/* returned. When unrecognized keystroke is entered, error message   */
/* appears and a FALSE is returned. Otherwise, the command code gets */
/* loaded and a TRUE is returned.                                    */
void display_cmds (int code)
{
	int
   	num_skipped_cmds,
		icmd, icol, irow, col_len,
		num_kb_cmds;
	KB_CMD
		*kb_cmd_list,
      *kb_cmd;

	get_menu (code, &kb_cmd_list);
	highvideo ();
	cprintf ("Keystroke Command List (Key Pkt-ID Description):");
	normvideo ();
	textcolor (LIGHTGRAY);
	textbackground (BLUE);
	clreol ();
   num_skipped_cmds = 0;
	for (icmd=0; icmd < MAX_KB_CMDS; icmd++)
	{
   	if (!tsipchat && kb_cmd_list[icmd].cmdcode == 0x11)
      {
		   num_skipped_cmds++;
      }
      else
      {
			if (!kb_cmd_list[icmd].cmdcode) break;
      }
	}
	if (icmd == MAX_KB_CMDS)
	{
		show_crlf();
		cprintf ("Menu too large or not terminated!");
	}
	num_kb_cmds = icmd;
	col_len = (num_kb_cmds + 2) / 3;
	for (irow = 0; irow < col_len; irow++) {
		show_crlf ();
		for (icol = 0; icol < 3; icol++) {
			icmd = irow + (col_len * icol) + num_skipped_cmds;
			if (icmd >= num_kb_cmds) break;
         kb_cmd = &kb_cmd_list[icmd];
		   if (kb_cmd->cmdcode == 0x10 || kb_cmd->cmdcode == 0x11)
         {
         	highvideo ();
	         if (kb_cmd->kbch <= cntrl('Z'))
   	      {
	     	   	cprintf (" ^%c", kb_cmd->kbch+CNTRLKEY);
	         }
   	      else
      	   {
         		cprintf ("  %c", kb_cmd->kbch);
	         }
     	   	cprintf ("  --  %s", kb_cmd->cmdtext);
			   normvideo ();
				textcolor (LIGHTGRAY);
				textbackground (BLUE);
         }
         else if (kb_cmd->kbch <= cntrl('Z'))
         {
         	cprintf (" ^%c 0x%2.2X %s", kb_cmd->kbch + CNTRLKEY,
            	kb_cmd->cmdcode, kb_cmd->cmdtext);
         }
         else
         {
         	cprintf ("  %c 0x%2.2X %s", kb_cmd->kbch,
            	kb_cmd->cmdcode, kb_cmd->cmdtext);
         }
		}
	}
	show_crlf ();
}

static void confirm_keystroke (int code, KB_CMD *kb_cmd)
{
	KB_CMD
		*kb_cmd_list;
	int
		kbch,
		ikb;

	kbch = kb_cmd->kbch;
	get_menu (code, &kb_cmd_list);
	for (ikb = 0; ikb < MAX_KB_CMDS && kb_cmd_list[ikb].cmdcode; ikb++) {
   	if (!tsipchat && kb_cmd_list[ikb].cmdcode == 0x11) continue;
		if (kbch == kb_cmd_list[ikb].kbch) {
			clreol ();
			*kb_cmd = kb_cmd_list[ikb];
         if (kb_cmd->kbch <= cntrl('Z'))
         {
         	cprintf (" '^%c' 0x%2.2X %s", kb_cmd->kbch + CNTRLKEY,
            	kb_cmd->cmdcode, kb_cmd->cmdtext);
         }
         else
         {
         	cprintf (" ' %c' 0x%2.2X %s", kb_cmd->kbch,
            	kb_cmd->cmdcode, kb_cmd->cmdtext);
         }
         if (kb_cmd->cmdtext[0] == '*')
         {
         	cprintf ("  (old form; newer command preferred)");
         }
			show_crlf ();
			return;
		}
	}
	if (ikb == MAX_KB_CMDS)
	{
		show_crlf();
		cprintf ("Menu too large or not terminated!");
	}
	if (isprint (kbch) || ((kbch >= cntrl('A')) && (kbch <= cntrl('Z'))))
   {
		cprintf ("Keystroke %c Unknown ", kbch);
	}
	if (((kbch >= cntrl('A')) && (kbch <= cntrl('Z'))))
   {
		cprintf ("Keystroke ^%c Unknown ", kbch - cntrl('A') + 'A');
	}
	clreol ();
	kb_cmd->kbch = 0x1B;
	kb_cmd->cmdcode = 0;
	kb_cmd->replycode = 0;
}

int DisplayMenuChoices (int code, KB_CMD *kb_cmd)
{
	start_menu_window ();
	display_cmds (code);
   highvideo ();
	cprintf ("Enter Keystroke: ");
   normvideo ();
   textcolor (LIGHTGRAY);
   textbackground (BLUE);
	kb_cmd->kbch = ProcessInputBytesWhileWaitingForKBHit ();
	confirm_keystroke (code, kb_cmd);
	end_menu_window ();
	return (kb_cmd->kbch == 0x1B);
}

/* returns TRUE if a command is to be sent to the receiver; otherwise   */
/* a FALSE is returned                                                  */
void ProcessTSIPKeystroke (TSIPPKT *pCmd, int kbch, int *reply)
{
	KB_CMD
		kb_cmd;

	num_replies = 1;
   pCmd->code = 0;
	set_cmd_esc (FALSE);
	if (kbch == '=')
   {
		DisplayMenuChoices (x8E_MENU, &kb_cmd);
   }
   else
   {
		kb_cmd.kbch = kbch;
		confirm_keystroke (ALL_MENU, &kb_cmd);
   }

	if (kb_cmd.kbch == 0x1B) {
		return;
	}
   kb_cmd.cmdfn (pCmd);
	if (cmd_esc ()) {
		show_crlf ();
		cprintf ("-- aborted --");
	}
	else
	{
		clreol ();
		/* general case; display the message if a report has been requested */
		if (kb_cmd.replycode) {
         if ((kb_cmd.replycode==0x010B || kb_cmd.replycode==0x01AD) && !num_replies)
         {
         	kb_cmd.replycode = 0x01A5;
            num_replies = 1;
         }
			reply[kb_cmd.replycode] = num_replies;
         if (num_replies)
         {
				switch (kb_cmd.replycode) {
				case 0x6D:
					reply [0x44] = 1;
					break;

				case 0x46:
					reply [0x4B] = 1;
					break;

				case 0x57:
					reply [0x42] = reply [0x83] = reply [0x4A] =
					reply [0x84] = reply [0x43] = reply [0x56] =
					reply [0x120] = reply [0x117] = reply [0x118] = 1;
					break;
				}
				if (kb_cmd.replycode > 0x100)
				{
					cprintf ("...waiting for reply (0x8F-%X)\r",
						kb_cmd.replycode - 0x100);
				}
				else
				{
					cprintf ("...waiting for reply (0x%X)\r",
						kb_cmd.replycode);
				}
         }
         else
         {
         	kb_cmd.replycode = 0;
         }
		}
	}
}


